"""Bayesian Generalized Linear Model implementation.

Implementation of Bayesian GLMs using a mixture of Gaussians posterior
approximation and auto-encoding variational Bayes inference. See [1]_ for the
posterior mixture idea, and [2]_ for the inference scheme.

.. [1] Gershman, S., Hoffman, M., & Blei, D. "Nonparametric variational
   inference". Proceedings of the international conference on machine learning.
   2012.
.. [2] Kingma, D. P., & Welling, M. "Auto-encoding variational Bayes".
   Proceedings of the 2nd International Conference on Learning Representations
   (ICLR). 2014.
"""
from __future__ import division

import numpy as np
import logging
from itertools import chain
from multiprocessing import Pool

from scipy.stats.distributions import gamma, norm
from scipy.optimize import brentq
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted, check_X_y, check_array
from sklearn.utils import check_random_state

from .utils import atleast_list, issequence
from .mathfun.special import logsumexp
from .basis_functions import apply_grad
from .optimize import sgd, structured_sgd, logtrick_sgd
from .btypes import Bound, Positive, Parameter


# Set up logging
log = logging.getLogger(__name__)


# Module settings
WGTRND = norm()  # Sampling distribution over mixture weights
COVRND = gamma(a=2, scale=0.5)  # Sampling distribution over mixture covariance
LOGITER = 500  # Number of SGD iterations between logging ELBO and hypers


class GeneralizedLinearModel(BaseEstimator, RegressorMixin):
    r"""
    Bayesian Generalized linear model (GLM).

    This provides a scikit learn compatible interface for the glm module.

    Parameters
    ----------
    likelihood : Object
        A likelihood object, see the likelihoods module.
    basis : Basis
        A basis object, see the basis_functions module.
    K : int, optional
        Number of diagonal Gaussian components to use to approximate the
        posterior distribution.
    maxiter : int, optional
        Maximum number of iterations of stochastic gradients to run.
    batch_size : int, optional
        number of observations to use per SGD batch.
    updater : SGDUpdater, optional
        The SGD learning rate updating algorithm to use, by default this is
        Adam. See revrand.optimize.sgd for different options.
    nsamples : int, optional
        Number of samples for sampling the expected likelihood and expected
        likelihood gradients
    nstarts : int, optional
        if there are any parameters with distributions as initial values, this
        determines how many random candidate starts shoulds be evaluated before
        commencing optimisation at the best candidate.
    random_state : None, int or RandomState, optional
        random seed

    Notes
    -----
    This approximates the posterior distribution over the weights with a
    mixture of Gaussians:

    .. math ::

        \mathbf{w} \sim \frac{1}{K} \sum^K_{k=1}
            \mathcal{N}(\mathbf{m_k}, \boldsymbol{\Psi}_k)

    where,

    .. math ::

        \boldsymbol{\Psi}_k = \text{diag}([\Psi_{k,1}, \ldots,
            \Psi_{k,D}]).

    This is so arbitrary likelihoods can be used with this algorithm, while
    still mainting flexible and tractable non-Gaussian posteriors. Additionaly
    this has the benefit that we have a reduced number of parameters to
    optimise (compared with full covariance Gaussians).

    The main differences between this implementation and the GLM in [1]_ are:
        - We use diagonal mixtures, as opposed to isotropic.
        - We use auto encoding variational Bayes (AEVB) inference [2]_ with
          stochastic gradients.

    This uses the python logging module for displaying learning status. To view
    these messages have something like,

    .. code ::

        import logging
        logging.basicConfig(level=logging.INFO)
        log = logging.getLogger(__name__)

    in your calling code.
    """

    def __init__(self,
                 likelihood,
                 basis,
                 K=10,
                 maxiter=3000,
                 batch_size=10,
                 updater=None,
                 nsamples=50,
                 nstarts=500,
                 random_state=None
                 ):
        """See class docstring."""
        self.likelihood = likelihood
        self.basis = basis
        self.K = K
        self.maxiter = maxiter
        self.batch_size = batch_size
        self.updater = updater
        self.nsamples = nsamples
        self.nstarts = nstarts
        self.random_state = random_state  # For clone compatibility
        self.random_ = check_random_state(self.random_state)

    def fit(self, X, y, likelihood_args=()):
        r"""
        Learn the parameters of a Bayesian generalized linear model (GLM).

        Parameters
        ----------
        X : ndarray
            (N, d) array input dataset (N samples, d dimensions).
        y : ndarray
            (N,) array targets (N samples)
        likelihood : Object
            A likelihood object, see the likelihoods module.
        likelihood_args : sequence, optional
            sequence of arguments to pass to the likelihood function. These are
            non-learnable parameters. They can be scalars or arrays of length
            N.
        """
        X, y = check_X_y(X, y)

        # Batch magnification factor
        N, _ = X.shape
        self.B_ = X.shape[0] / self.batch_size
        self.D_ = self.basis.get_dim(X)

        # Pack data
        likelihood_args = _reshape_likelihood_args(likelihood_args, N)
        data = (X, y) + likelihood_args

        # Pack params
        params = [Parameter(WGTRND, Bound(), shape=(self.D_, self.K)),
                  Parameter(COVRND, Positive(), shape=(self.D_, self.K)),
                  self.basis.regularizer,
                  self.likelihood.params,
                  self.basis.params]

        log.info("Optimising parameters...")
        self.__it = -self.nstarts  # Keeping track of iterations for logging
        nsgd = structured_sgd(logtrick_sgd(sgd))
        res = nsgd(self._elbo,
                   params,
                   data,
                   eval_obj=True,
                   maxiter=self.maxiter,
                   updater=self.updater,
                   batch_size=self.batch_size,
                   random_state=self.random_,
                   nstarts=self.nstarts
                   )

        # Unpack params
        (self.weights_,
         self.covariance_,
         self.regularizer_,
         self.like_hypers_,
         self.basis_hypers_
         ) = res.x

        log.info("Finished! reg = {}, likelihood_hypers = {}, "
                 "basis_hypers = {}, message: {}."
                 .format(self.regularizer_,
                         self.like_hypers_,
                         self.basis_hypers_,
                         res.message))

        return self

    def _elbo(self, m, C, reg, lpars, bpars, X, y, *largs):
        # Full evidence lower bound objective with AEVB

        # Shapes
        D, K = m.shape

        # Make sure hypers and args can be unpacked into callables
        largs = tuple(chain(atleast_list(lpars), largs))

        # Basis function
        Phi = self.basis.transform(X, *atleast_list(bpars))  # M x D

        # Get regularizer
        L, slices = self.basis.regularizer_diagonal(X, *atleast_list(reg))
        iL = 1. / L[:, np.newaxis]

        # Posterior entropy lower bound terms
        logNkl = _qmatrix(m, C)
        logzk = logsumexp(logNkl, axis=0)

        # Preallocate variational parameter gradients and ELL
        dm = np.empty_like(m)
        dC = np.empty_like(C)
        Ell = np.empty(K, dtype=float)

        # Zero starts for sums over posterior mixtures
        dlpars = [np.zeros_like(p) for p in atleast_list(lpars)]
        EdPhi = np.zeros_like(Phi)

        # Log status, only do this occasionally to save cpu
        dolog = (self.__it % LOGITER == 0) or (self.__it == self.maxiter - 1)

        # Only calculate ELBO when sampling parameters (__it < 0) or logging
        calc_ll = dolog or (self.__it < 0)

        # Big loop though posterior mixtures for calculating stuff
        for k in range(K):

            # Sample expected likelihood and gradients
            Edmk, EdCk, EdPhik, Edlpars, Ell[k] = \
                self._reparam_k(m[:, k], C[:, k], y, Phi, largs, calc_ll)
            EdPhi += EdPhik / K

            # Weight factors for each component in the gradients
            Nkl_zk = np.exp(logNkl[:, k] - logzk[k])
            Nkl_zl = np.exp(logNkl[:, k] - logzk)
            alpha = (Nkl_zk + Nkl_zl)

            # Posterior mean and covariance gradients
            mkmj = m[:, k][:, np.newaxis] - m
            iCkCj = 1. / (C[:, k][:, np.newaxis] + C)

            dm[:, k] = (self.B_ * Edmk - m[:, k] / L
                        + (iCkCj * mkmj).dot(alpha)) / K
            dC[:, k] = (self.B_ * EdCk - 1. / L
                        + (iCkCj - (mkmj * iCkCj)**2).dot(alpha)) / (2 * K)

            # Likelihood parameter gradients
            for i, Edlpar in enumerate(Edlpars):
                dlpars[i] -= Edlpar / K

        # Regularizer gradient
        def dreg(s):
            dL = 0.5 * (((m[s]**2 + C[s]) * iL[s]**2).sum() / K - iL[s].sum())
            return -dL

        dL = list(map(dreg, slices)) if issequence(slices) else dreg(slices)

        # Basis function parameter gradients
        dtheta = lambda dPhi: -(EdPhi * dPhi).sum()
        dbpars = apply_grad(dtheta, self.basis.grad(X, *atleast_list(bpars)))

        # Approximate evidence lower bound
        ELBO = -np.inf
        if calc_ll:
            ELBO = (Ell.sum() * self.B_
                    - 0.5 * D * K * np.log(2 * np.pi)
                    - 0.5 * K * np.log(L).sum()
                    - 0.5 * ((m**2 + C) * iL).sum()
                    - logzk.sum() + np.log(K)) / K

        if dolog:
            rs_mes = "Random starts: " if self.__it < 0 else ""
            log.info("{}Iter {}: ELBO = {}, reg = {}, like_hypers = {}, "
                     "basis_hypers = {}"
                     .format(rs_mes, self.__it, ELBO, reg, lpars, bpars))

        self.__it += 1

        return -ELBO, [-dm, -dC, dL, dlpars, dbpars]

    def _reparam_k(self, mk, Ck, y, Phi, largs, calc_ll=True):
        # AEVB's reparameterisation trick

        # Sample the latent function and its derivative
        e = self.random_.randn(self.nsamples, len(mk))
        Sk = np.sqrt(Ck)
        ws = mk + Sk * e  # L x D
        fs = ws.dot(Phi.T)  # L x M
        dfs = self.likelihood.df(y, fs, *largs)  # L x M

        # Expected gradients
        Edws = dfs.dot(Phi)  # L x D, dweight samples
        Edm = Edws.sum(axis=0) / self.nsamples  # D
        EdC = (Edws * e / Sk).sum(axis=0) / self.nsamples  # D
        EdPhi = dfs.T.dot(ws) / self.nsamples  # M x D

        # Structured likelihood parameter gradients
        Edlpars = atleast_list(self.likelihood.dp(y, fs, *largs))
        for i, Edlpar in enumerate(Edlpars):
            Edlpars[i] = Edlpar.sum() / self.nsamples

        # Expected ll
        Ell = np.inf
        if calc_ll:
            Ell = self.likelihood.loglike(y, fs, *largs).sum() / self.nsamples

        return Edm, EdC, EdPhi, Edlpars, Ell

    def predict(self, X, nsamples=200, likelihood_args=()):
        """
        Predict target values from Bayesian generalized linear regression.

        Parameters
        ----------
        X : ndarray
            (N*,d) array query input dataset (N* samples, d dimensions).
        nsamples : int, optional
            Number of samples for sampling the expected target values from the
            predictive distribution.
        likelihood_args : sequence, optional
            sequence of arguments to pass to the likelihood function. These are
            non-learnable parameters. They can be scalars or arrays of length
            N.

        Returns
        -------
        Ey : ndarray
            The expected value of y* for the query inputs, X* of shape (N*,).
        """
        Ey, _ = self.predict_moments(X, nsamples, likelihood_args)

        return Ey

    def predict_moments(self, X, nsamples=200, likelihood_args=()):
        r"""
        Predictive moments, in particular mean and variance, of a Bayesian GLM.

        This function uses Monte-Carlo sampling to evaluate the predictive mean
        and variance of a Bayesian GLM. The exact expressions evaluated are,

        .. math ::

            \mathbb{E}[y^* | \mathbf{x^*}, \mathbf{X}, y] &=
                \int \mathbb{E}[y^* | \mathbf{w}, \phi(\mathbf{x}^*)]
                p(\mathbf{w} | \mathbf{y}, \boldsymbol\Phi) d\mathbf{w},

            \mathbb{V}[y^* | \mathbf{x^*}, \mathbf{X}, y] &=
                \int \left(\mathbb{E}[y^* | \mathbf{w}, \phi(\mathbf{x}^*)]
                - \mathbb{E}[y^* | \mathbf{x^*}, \mathbf{X}, y]\right)^2
                p(\mathbf{w} | \mathbf{y}, \boldsymbol\Phi) d\mathbf{w},

        where :math:`\mathbb{E}[y^* | \mathbf{w}, \phi(\mathbf{x}^*)]` is the
        the expected value of :math:`y^*` from  the likelihood, and
        :math:`p(\mathbf{w} | \mathbf{y}, \boldsymbol\Phi)` is the posterior
        distribution over weights (from ``learn``). Here are few concrete
        examples of how we can use these values,

        - Gaussian likelihood: these are just the predicted mean and variance,
          see ``revrand.regression.predict``
        - Bernoulli likelihood: The expected value is the probability,
          :math:`p(y^* = 1)`, i.e. the probability of class one. The variance
          may not be so useful.
        - Poisson likelihood: The expected value is similar conceptually to the
          Gaussian case, and is also a *continuous* value. The median (50%
          quantile) from ``predict_interval`` is a discrete value. Again,
          the variance in this instance may not be so useful.

        Parameters
        ----------
        X : ndarray
            (N*,d) array query input dataset (N* samples, d
            dimensions).
        nsamples : int, optional
            Number of samples for sampling the expected moments from the
            predictive distribution.
        likelihood_args : sequence, optional
            sequence of arguments to pass to the likelihood function. These are
            non-learnable parameters. They can be scalars or arrays of length
            N.

        Returns
        -------
        Ey : ndarray
            The expected value of y* for the query inputs, X* of shape (N*,).
        Vy : ndarray
            The expected variance of y* (excluding likelihood noise terms) for
            the query inputs, X* of shape (N*,).
        """
        # Get latent function samples
        N = X.shape[0]
        ys = np.empty((N, nsamples))
        fsamples = self._sample_func(X, nsamples)

        # Push samples though likelihood expected value
        Eyargs = tuple(chain(atleast_list(self.like_hypers_), likelihood_args))
        for i, f in enumerate(fsamples):
            ys[:, i] = self.likelihood.Ey(f, *Eyargs)

        # Average transformed samples (MC integration)
        Ey = ys.mean(axis=1)
        Vy = ((ys - Ey[:, np.newaxis])**2).mean(axis=1)

        return Ey, Vy

    def predict_logpdf(self, X, y, nsamples=200, likelihood_args=()):
        r"""
        Predictive log-probability density function of a Bayesian GLM.

        Parameters
        ----------
        X : ndarray
            (N*,d) array query input dataset (N* samples, D dimensions).
        y : float or ndarray
            The test observations of shape (N*,) to evaluate under,
            :math:`\log p(y^* |\mathbf{x}^*, \mathbf{X}, y)`.
        nsamples : int, optional
            Number of samples for sampling the log predictive distribution.
        likelihood_args : sequence, optional
            sequence of arguments to pass to the likelihood function. These are
            non-learnable parameters. They can be scalars or arrays of length
            N*.

        Returns
        -------
        logp : ndarray
           The log probability of y* given X* of shape (N*,).
        logp_min : ndarray
            The minimum sampled values of the predicted log probability (same
            shape as p)
        logp_max : ndarray
            The maximum sampled values of the predicted log probability (same
            shape as p)
        """
        X, y = check_X_y(X, y)

        # Get latent function samples
        N = X.shape[0]
        ps = np.empty((N, nsamples))
        fsamples = self._sample_func(X, nsamples)

        # Push samples though likelihood pdf
        llargs = tuple(chain(atleast_list(self.like_hypers_), likelihood_args))
        for i, f in enumerate(fsamples):
            ps[:, i] = self.likelihood.loglike(y, f, *llargs)

        # Average transformed samples (MC integration)
        logp = ps.mean(axis=1)
        logp_min = ps.min(axis=1)
        logp_max = ps.max(axis=1)

        return logp, logp_min, logp_max

    def predict_cdf(self, X, quantile, nsamples=200, likelihood_args=()):
        r"""
        Predictive cumulative density function of a Bayesian GLM.

        Parameters
        ----------
        X : ndarray
            (N*,d) array query input dataset (N* samples, D dimensions).
        quantile : float
            The predictive probability, :math:`p(y^* \leq \text{quantile} |
            \mathbf{x}^*, \mathbf{X}, y)`.
        nsamples : int, optional
            Number of samples for sampling the predictive CDF.
        likelihood_args : sequence, optional
            sequence of arguments to pass to the likelihood function. These are
            non-learnable parameters. They can be scalars or arrays of length
            N*.
        nsamples : int, optional
            The number of samples to draw from the posterior in order to
            approximate the predictive mean and variance.

        Returns
        -------
        p : ndarray
           The probability of y* <= quantile for the query inputs, X* of shape
           (N*,).
        p_min : ndarray
            The minimum sampled values of the predicted probability (same shape
            as p)
        p_max : ndarray
            The maximum sampled values of the predicted probability (same shape
            as p)
        """
        # Get latent function samples
        N = X.shape[0]
        ps = np.empty((N, nsamples))
        fsamples = self._sample_func(X, nsamples)

        # Push samples though likelihood cdf
        cdfarg = tuple(chain(atleast_list(self.like_hypers_), likelihood_args))
        for i, f in enumerate(fsamples):
            ps[:, i] = self.likelihood.cdf(quantile, f, *cdfarg)

        # Average transformed samples (MC integration)
        p = ps.mean(axis=1)
        p_min = ps.min(axis=1)
        p_max = ps.max(axis=1)

        return p, p_min, p_max

    def predict_interval(self, X, percentile, nsamples=200, likelihood_args=(),
                         multiproc=True):
        """
        Predictive percentile interval (upper and lower quantiles).

        Parameters
        ----------
        X : ndarray
            (N*,d) array query input dataset (N* samples, D dimensions).
        percentile : float
            The percentile confidence interval (e.g. 95%) to return.
        nsamples : int, optional
            Number of samples for sampling the predictive percentiles.
        likelihood_args : sequence, optional
            sequence of arguments to pass to the likelihood function. These are
            non-learnable parameters. They can be scalars or arrays of length
            N*.
        multiproc : bool, optional
            Use multiprocessing to paralellise this prediction computation.

        Returns
        -------
        ql : ndarray
            The lower end point of the interval with shape (N*,)
        qu : ndarray
            The upper end point of the interval with shape (N*,)
        """
        N = X.shape[0]

        # Generate latent function samples per observation (n in N)
        fsamples = self._sample_func(X, nsamples, genaxis=0)

        # Make sure likelihood_args is consistent with work
        if len(likelihood_args) > 0:
            likelihood_args = _reshape_likelihood_args(likelihood_args, N)

        # Now create work for distrbuted workers
        like_hypers = atleast_list(self.like_hypers_)
        work = ((f[0], self.likelihood, like_hypers, f[1:], percentile)
                for f in zip(fsamples, *likelihood_args))

        # Distribute sampling and rootfinding
        if multiproc:
            pool = Pool()
            res = pool.map(_star_rootfinding, work)
            pool.close()
            pool.join()
        else:
            res = [_rootfinding(*w) for w in work]

        # Get results of work
        ql, qu = zip(*res)
        return np.array(ql), np.array(qu)

    def _sample_func(self, X, nsamples, genaxis=1):
        """
        Generate samples from the posterior latent function mixtures of the GLM
        for query inputs, X*.

        Parameters
        ----------
        X : ndarray
            (N*, d) array query input dataset (N* samples, D dimensions).
        nsamples : int
            Number of samples for sampling the latent function.
        genaxis : int
            Axis to return samples from, i.e.
            - ``genaxis=1`` will give you one sample at a time of f for ALL
                observations (so it will iterate over nsamples).
            - ``genaxis=0`` will give you all samples of f for ONE
                observation at a time (so it will iterate through X*, row by
                row)

        Yields
        ------
        fsamples : ndarray
            of shape (N*,) if ``genaxis=1`` with each call being a sample
            from the mixture of latent functions over all N*. Or of shape
            (nsamples,) if ``genaxis=0``, with each call being a all
            samples for an observation, n in N*.
        """
        check_is_fitted(self, ['weights_', 'covariance_', 'basis_hypers_',
                               'like_hypers_', 'regularizer_'])
        X = check_array(X)
        D, K = self.weights_.shape

        # Generate weight samples from all mixture components
        k = self.random_.randint(0, K, size=(nsamples,))
        w = self.weights_[:, k] + self.random_.randn(D, nsamples) \
            * np.sqrt(self.covariance_[:, k])

        # Do this here for *massive* speed improvements
        Phi = self.basis.transform(X, *atleast_list(self.basis_hypers_))

        # Now generate latent functions samples either colwise or rowwise
        if genaxis == 1:
            fs = (Phi.dot(ws) for ws in w.T)
        elif genaxis == 0:
            fs = (phi_n.dot(w) for phi_n in Phi)
        else:
            raise ValueError("Invalid axis to generate samples from")

        return fs

    def __repr__(self):
        """Representation."""
        return "{}(likelihood={}, basis={}, K={}, maxiter={}, batch_size={},"\
            "updater={}, nsamples={}, nstarts={}, random_state={})".format(
                self.__class__.__name__,
                self.likelihood,
                self.basis,
                self.K,
                self.maxiter,
                self.batch_size,
                self.updater,
                self.nsamples,
                self.nstarts,
                self.random_state
            )


# For GB/AU spelling
class GeneralisedLinearModel(GeneralizedLinearModel):
    pass


#
#  Internal Module Utilities
#

def _reshape_likelihood_args(likelihood_args, N):

    reshape_args = []
    for l in likelihood_args:
        if np.isscalar(l):
            l = l * np.ones(N)

        if (np.shape(l)[0] != N) and (len(l) != 0):
            raise ValueError("Likelihood arguments not a compatible shape!")

        reshape_args.append(l)

    return tuple(reshape_args)


# For python 2.7 compatibility instead of pool.starmap
def _star_rootfinding(args):

    return _rootfinding(*args)


def _rootfinding(fn, likelihood, likelihood_hypers, likelihood_args,
                 percentile):

    # CDF minus percentile for quantile root finding
    predCDF = lambda q, fs, percent: \
        (likelihood.cdf(q, fs, *chain(likelihood_hypers,
                                      likelihood_args))).mean() - percent

    # Convert alpha into percentages and get (conservative) bounds for brentq
    lpercent = (1 - percentile) / 2
    upercent = 1 - lpercent
    Eyn = likelihood.Ey(fn, *chain(likelihood_hypers, likelihood_args)).mean()
    lb, ub = -1000 * max(Eyn, 1), 1000 * max(Eyn, 1)

    # Do the root finding optimisation for upper and lower quantiles
    try:
        qln = brentq(predCDF, a=lb, b=ub, args=(fn, lpercent))
    except ValueError:
        qln = np.nan

    try:
        qun = brentq(predCDF, a=lb, b=ub, args=(fn, upercent))
    except ValueError:
        qun = np.nan

    return qln, qun


def _dgausll(x, mean, dcov):
    # This is faster than calling scipy.stats.norm.logpdf

    D = len(x)
    return - 0.5 * (D * np.log(2 * np.pi) + np.log(dcov).sum()
                    + ((x - mean)**2 / dcov).sum())


def _qmatrix(m, C):

    K = m.shape[1]
    logq = [[_dgausll(m[:, i], m[:, j], C[:, i] + C[:, j])
             for i in range(K)]
            for j in range(K)]

    return np.array(logq)



"""
The standard Bayesian linear regression model.

By using the appropriate bases, this will also yield an implementation of the
"A la Carte" GP [1]_.

.. [1] Yang, Z., Smola, A. J., Song, L., & Wilson, A. G. "A la Carte --
   Learning Fast Kernels". Proceedings of the Eighteenth International
   Conference on Artificial Intelligence and Statistics, pp. 1098-1106,
   2015.
"""

from __future__ import division

import logging
from functools import partial

import numpy as np
from scipy.stats import gamma
from scipy.optimize import minimize
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted, check_X_y, check_array
from sklearn.utils import check_random_state

from .utils import atleast_list, issequence
from .mathfun.linalg import solve_posdef
from .optimize import structured_minimizer, logtrick_minimizer
from .btypes import Parameter, Positive
from .basis_functions import apply_grad

# Set up logging
log = logging.getLogger(__name__)


class StandardLinearModel(BaseEstimator, RegressorMixin):
    """
    Bayesian Standard linear model.

    Parameters
    ----------
    basis : Basis
        A basis object, see the basis_functions module.
    var : Parameter, optional
        observation variance initial value.
    tol : float, optional
        optimiser function tolerance convergence criterion.
    maxiter : int, optional
        maximum number of iterations for the optimiser.
    nstarts : int, optional
        if there are any parameters with distributions as initial values, this
        determines how many random candidate starts shoulds be evaluated before
        commencing optimisation at the best candidate.
    random_state : None, int or RandomState, optional
        random seed (mainly for random starts)
    """

    def __init__(self,
                 basis,
                 var=Parameter(gamma(1.), Positive()),
                 tol=1e-8,
                 maxiter=1000,
                 nstarts=100,
                 random_state=None
                 ):
        """See class docstring."""
        self.basis = basis
        self.var = var
        self.tol = tol
        self.maxiter = maxiter
        self.nstarts = nstarts
        self.random_state = random_state
        self.random_ = check_random_state(random_state)

    def fit(self, X, y):
        """
        Learn the hyperparameters of a Bayesian linear regressor.

        Parameters
        ----------
        X : ndarray
            (N, d) array input dataset (N samples, d dimensions).
        y : ndarray
            (N,) array targets (N samples)

        Returns
        -------
        self

        Notes
        -----
        This actually optimises the evidence lower bound on log marginal
        likelihood, rather than log marginal likelihood directly. In the case
        of a full posterior convariance matrix, this bound is tight and the
        exact solution will be found (modulo local minima for the
        hyperparameters).

        This uses the python logging module for displaying learning status. To
        view these messages have something like,

        .. code ::

            import logging
            logging.basicConfig(level=logging.INFO)
            log = logging.getLogger(__name__)

        in your calling code.
        """
        X, y = check_X_y(X, y)

        self.obj_ = -np.inf

        # Make list of parameters and decorate optimiser to undestand this
        params = [self.var, self.basis.regularizer, self.basis.params]
        nmin = structured_minimizer(logtrick_minimizer(minimize))

        # Close over objective and learn parameters
        elbo = partial(StandardLinearModel._elbo, self, X, y)
        res = nmin(elbo,
                   params,
                   method='L-BFGS-B',
                   jac=True,
                   tol=self.tol,
                   options={'maxiter': self.maxiter, 'maxcor': 100},
                   random_state=self.random_,
                   nstarts=self.nstarts
                   )

        # Upack learned parameters and report
        self.var_, self.regularizer_, self.hypers_ = res.x

        log.info("Done! ELBO = {}, var = {}, reg = {}, hypers = {}, "
                 "message = {}."
                 .format(-res['fun'],
                         self.var_,
                         self.regularizer_,
                         self.hypers_,
                         res.message)
                 )

        return self

    def _elbo(self, X, y, var, reg, hypers):

        # Get Basis
        Phi = self.basis.transform(X, *atleast_list(hypers))  # N x D
        PhiPhi = Phi.T.dot(Phi)
        N, D = Phi.shape

        # Get regularizer
        L, slices = self.basis.regularizer_diagonal(X, *atleast_list(reg))
        iL = 1. / L

        # Posterior Parameters
        iC = np.diag(iL) + PhiPhi / var
        C, logdetiC = solve_posdef(iC, np.eye(D))
        logdetC = - logdetiC
        m = C.dot(Phi.T.dot(y)) / var

        # Common calcs
        TrPhiPhiC = (PhiPhi * C).sum()
        Err = y - Phi.dot(m)
        sqErr = (Err**2).sum()

        # Calculate ELBO
        ELBO = -0.5 * (N * np.log(2 * np.pi * var)
                       + sqErr / var
                       + TrPhiPhiC / var
                       + ((m**2 + C.diagonal()) * iL).sum()
                       - logdetC
                       + np.log(L).sum()
                       - D)

        # Cache optimal parameters so we don't have to recompute them later
        if ELBO > self.obj_:
            self.weights_ = m
            self.covariance_ = C
            self.obj_ = ELBO

        log.info("ELBO = {}, var = {}, reg = {}, bparams = {}."
                 .format(ELBO, var, reg, hypers))

        # Grad var
        dvar = 0.5 * (-N + (sqErr + TrPhiPhiC) / var) / var

        # Grad reg
        def dreg(s):
            return - 0.5 * (((m[s]**2 + C[s, s].diagonal()) * iL[s]**2).sum()
                            - iL[s].sum())

        dL = list(map(dreg, slices)) if issequence(slices) else dreg(slices)

        # Get structured basis function gradients
        def dhyps(dPhi):
            return - (m.T.dot(Err.dot(dPhi))
                      - (dPhi.T.dot(Phi) * C).sum()) / var

        dhypers = apply_grad(dhyps, self.basis.grad(X, *atleast_list(hypers)))

        return -ELBO, [-dvar, dL, dhypers]

    def predict(self, X):
        """
        Predict mean from Bayesian linear regression.

        Parameters
        ----------
        X : ndarray
            (N*,d) array query input dataset (N* samples, d dimensions).

        Returns
        -------
        Ey : ndarray
            The expected value of y* for the query inputs, X* of shape (N*,).
        """
        Ey, _ = self.predict_moments(X)

        return Ey

    def predict_moments(self, X):
        """
        Full predictive distribution from Bayesian linear regression.

        Parameters
        ----------
        X : ndarray
            (N*,d) array query input dataset (N* samples, d dimensions).

        Returns
        -------
        Ey : ndarray
            The expected value of y* for the query inputs, X* of shape (N*,).
        Vy : ndarray
            The expected variance of y* for the query inputs, X* of shape
            (N*,).
        """
        check_is_fitted(self, ['var_', 'regularizer_', 'weights_',
                               'covariance_', 'hypers_'])
        X = check_array(X)

        Phi = self.basis.transform(X, *atleast_list(self.hypers_))
        Ey = Phi.dot(self.weights_)
        Vf = (Phi.dot(self.covariance_) * Phi).sum(axis=1)

        return Ey, Vf + self.var_

    def __repr__(self):
        """Representation."""
        return "{}(basis={}, var={}, tol={}, maxiter={}, nstarts={}, "\
            "random_state={})".format(
                self.__class__.__name__,
                self.basis,
                self.var,
                self.tol,
                self.maxiter,
                self.nstarts,
                self.random_state
            )



#!/usr/bin/env python
"""Setup utility for the revrand package."""

from setuptools import setup, find_packages

from setuptools.command.test import test as TestCommand


class PyTest(TestCommand, object):

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        super(PyTest, self).initialize_options()
        self.pytest_args = []

    def finalize_options(self):
        super(PyTest, self).finalize_options()
        self.test_suite = True
        self.test_args = []

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        exit(pytest.main(self.pytest_args))


setup(
    name='revrand',
    version='0.9.10',
    description='A library of scalable Bayesian generalised linear models with'
                ' fancy features',
    author='Daniel Steinberg',
    author_email='daniel.steinberg@nicta.com.au',
    url='http://github.com/nicta/revrand',
    packages=find_packages(),
    cmdclass={
        'test': PyTest
    },
    tests_require=['pytest'],
    install_requires=[
        'scipy >= 0.15.1',
        'numpy >= 1.8.2',
        'six >= 1.9.0',
        'decorator >= 4.0.6',
        'scikit-learn >= 0.18.0'
    ],
    extras_require={
        'demos': [
            'unipath',
            'requests',
            'matplotlib',
        ],
    },
    license="Apache Software License 2.0",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis"
    ]
)



import numpy as np
import numbers

from sklearn.utils.validation import check_random_state


def endless_permutations(N, random_state=None):
    """
    Generate an endless sequence of random integers from permutations of the
    set [0, ..., N).

    If we call this N times, we will sweep through the entire set without
    replacement, on the (N+1)th call a new permutation will be created, etc.

    Parameters
    ----------
    N: int
        the length of the set
    random_state: int or RandomState, optional
        random seed

    Yields
    ------
    int:
        a random int from the set [0, ..., N)
    """
    generator = check_random_state(random_state)
    while True:
        batch_inds = generator.permutation(N)
        for b in batch_inds:
            yield b



"""Bound types and bounded parameter types."""

from collections import namedtuple
from itertools import chain

import numpy as np
from sklearn.utils import check_random_state


class _BoundMixin(object):
    """Methods for the bound objects."""

    def check(self, value):
        """
        Check a value falls within a bound.

        Parameters
        ----------
        value : scalar or ndarray
            value to test

        Returns
        -------
        bool:
            If all values fall within bounds

        Example
        -------
        >>> bnd = Bound(1, 2)
        >>> bnd.check(1.5)
        True
        >>> bnd.check(3)
        False
        >>> bnd.check(np.ones(10))
        True
        >>> bnd.check(np.array([1, 3, 1.5]))
        False
        """
        if self.lower:
            if np.any(value < self.lower):
                return False

        if self.upper:
            if np.any(value > self.upper):
                return False

        return True

    def clip(self, value):
        """
        Clip a value to a bound.

        Parameters
        ----------
        value : scalar or ndarray
            value to clip

        Returns
        -------
        scalar or ndarray :
            of the same shape as value, bit with each element clipped to fall
            within the specified bounds

        Example
        -------
        >>> bnd = Bound(1, 2)
        >>> bnd.clip(1.5)
        1.5
        >>> bnd.clip(3)
        2
        >>> bnd.clip(np.array([1, 3, 1.5]))
        array([ 1. ,  2. ,  1.5])
        >>> bnd = Bound(None, None)
        >>> bnd.clip(np.array([1, 3, 1.5]))
        array([ 1. ,  3. ,  1.5])
        """
        if not self.lower and not self.upper:
            return value

        return np.clip(value, self.lower, self.upper)


class Bound(namedtuple('Bound', ['lower', 'upper']), _BoundMixin):
    """
    Define bounds on a variable for the optimiser.

    This defaults to all real values allowed (i.e. no bounds).

    Parameters
    ----------
    lower : float
        The lower bound.
    upper : float
        The upper bound.

    Attributes
    ----------
    lower : float
        The lower bound.
    upper : float
        The upper bound.

    Examples
    --------
    >>> b = Bound(1e-10, upper=1e-5)
    >>> b
    Bound(lower=1e-10, upper=1e-05)
    >>> b.lower
    1e-10
    >>> b.upper
    1e-05
    >>> isinstance(b, tuple)
    True
    >>> tuple(b)
    (1e-10, 1e-05)
    >>> lower, upper = b
    >>> lower
    1e-10
    >>> upper
    1e-05
    >>> Bound(42, 10)
    Traceback (most recent call last):
        ...
    ValueError: lower bound cannot be greater than upper bound!
    """

    def __new__(cls, lower=None, upper=None):
        """See class docstring."""
        if lower is not None and upper is not None:
            if lower > upper:
                raise ValueError('lower bound cannot be greater than upper '
                                 'bound!')
        return super(Bound, cls).__new__(cls, lower, upper)

    def __getnewargs__(self):
        """Required for pickling."""
        return (self.lower, self.upper)

    def __repr__(self):
        """Representation."""
        return "{}(lower={}, upper={})".format(
            self.__class__.__name__,
            self.lower,
            self.upper
        )


class Positive(namedtuple('Positive', ['lower', 'upper']), _BoundMixin):
    """
    Define a positive only bound for the optimiser.

    This may induce the 'log trick' in the optimiser (when using an appropriate
    decorator), which will ignore the 'smallest' value (but will stay above 0).

    Parameters
    ---------
    upper : float
        The largest value allowed for the optimiser to evaluate (if not using
        the log trick).

    Examples
    --------
    >>> b = Positive()
    >>> b
    Positive(upper=None)

    Since ``tuple`` (and by extension its descendents) are immutable,
    the lower bound for all instances of ``Positive`` are guaranteed to
    be positive.
    """

    def __new__(cls, upper=None):
        """See class docstring."""
        lower = 1e-14
        if upper is not None:
            if lower > upper:
                raise ValueError('Upper bound must be greater than {}'
                                 .format(lower))

        return super(Positive, cls).__new__(cls, lower, upper)

    def __getnewargs__(self):
        """Required for pickling."""
        return (self.upper,)

    def __repr__(self):
        """Representation."""
        return "{}(upper={})".format(self.__class__.__name__, self.upper)


class Parameter(object):
    """
    A Parameter class that associates a value with a bound.

    Attributes
    ----------
    value : None, scalar, ndarray, scipy.stats, optional
        a value or distribution to associate with this parameter. This is
        typically used as an initial value for an optimizer, and if a random
        starts optimiser is used (eg. revrand.optimize.structured_minimizer) it
        can draw randomly from the distribution.
    bound : Bound, optional
        a Bound tuple that describes the valid range for all of the elements in
        value
    shape : tuple, optional
        the shape of value, this is ignored if value is not a scipy.stats
        distribution (i.e. it is automatically determined).

    Note
    ----
    - If ``value`` is initialised as a distribution from ``scipy.stats``, then
      ``self.value`` is actually the mean of the distribution.
    - If you want to set value to a ``scipy.stats`` distribution, and also
      associate it with a ``shape`` (i.e. you want an ``ndarray`` random
      variable), then also set the ``shape`` parameters to the desired
      dimensions.

    Examples
    --------
    Null

    >>> p = Parameter()
    >>> p.value
    []
    >>> p.has_value
    False

    Scalar

    >>> p = Parameter(1.2, Bound(1, 2))
    >>> p.shape
    ()
    >>> p.value
    1.2
    >>> p.rvs()
    1.2
    >>> p.is_random
    False

    ndarray

    >>> p = Parameter(np.ones(3), Positive())
    >>> p.shape
    (3,)
    >>> p.value
    array([ 1.,  1.,  1.])
    >>> p.rvs()
    array([ 1.,  1.,  1.])

    ``scipy.stats`` scalar

    >>> from scipy.stats import gamma
    >>> p = Parameter(gamma(a=1, scale=1), Positive())
    >>> p.value == gamma(a=1, scale=1).mean()
    True
    >>> np.isscalar(p.rvs())
    True
    >>> p.is_random
    True

    ``scipy.stats`` ndarray

    >>> from scipy.stats import gamma
    >>> p = Parameter(gamma(a=1, scale=1), Positive(), shape=(3,))
    >>> all(p.value == np.ones(3) * gamma(a=1, scale=1).mean())
    True
    >>> p.rvs().shape == (3,)
    True
    """

    def __init__(self, value=None, bounds=Bound(), shape=()):
        """See class docstring."""
        value = [] if value is None else value

        if not hasattr(value, 'rvs'):
            if np.any(value) and not bounds.check(value):
                raise ValueError("Value not within bounds!")

            self.value = value
            self.dist = None
            self.shape = np.shape(value)
        else:
            self.dist = value
            self.shape = shape
            mean = bounds.clip(self.dist.mean())
            self.value = mean if shape == () else mean * np.ones(shape)

        self.bounds = bounds

    def rvs(self, random_state=None):
        r"""
        Draw a random value from this Parameter's distribution.

        If ``value`` was not initialised with a ``scipy.stats`` object, then
        the scalar/ndarray value is returned.

        Parameters
        ----------
        random_state : None, int or RandomState, optional
            random seed

        Returns
        -------
        ndarray :
            of size ``self.shape``, a random draw from the distribution, or
            ``self.value`` if not initialised with a ``scipy.stats`` object.

        Note
        ----
        Random draws are *clipped* to the bounds, and so it is up to the user
        to input a sensible sampling distribution!
        """
        # No sampling distibution
        if self.dist is None:
            return self.value

        # Unconstrained samples
        rs = check_random_state(random_state)
        samples = self.dist.rvs(size=self.shape, random_state=rs)

        # Bound the samples
        samples = self.bounds.clip(samples)

        return samples

    @property
    def has_value(self):
        """Test if this Parameter has a value, or is "null"."""
        return self.shape != (0,)

    @property
    def is_random(self):
        """Test if this Parameter was initialised with a distribution."""
        return self.dist is not None

    @property
    def is_scalar(self):
        """Test if the Parameter is for a scalar value."""
        return self.has_value and self.shape == ()

    def __repr__(self):
        """Representation."""
        return "{}(value={}, bounds={}, shape={})".format(
            self.__class__.__name__,
            self.value,
            self.bounds,
            self.shape
        )


def ravel(parameter, random_state=None):
    """
    Flatten a ``Parameter``.

    Parameters
    ----------
    parameter: Parameter
        A ``Parameter`` object

    Returns
    -------
    flatvalue: ndarray
        a flattened array of shape ``(prod(parameter.shape),)``
    flatbounds: list
        a list of bound tuples of length ``prod(parameter.shape)``
    """
    flatvalue = np.ravel(parameter.rvs(random_state=random_state))
    flatbounds = [parameter.bounds
                  for _ in range(np.prod(parameter.shape, dtype=int))]

    return flatvalue, flatbounds


def hstack(tup):
    """
    Horizontally stack a sequence of value bounds pairs.

    Parameters
    ----------
    tup: sequence
        a sequence of value, ``Bound`` pairs

    Returns
    -------
    value: ndarray
        a horizontally concatenated array1d
    bounds:
        a list of Bounds
    """
    vals, bounds = zip(*tup)
    stackvalue = np.hstack(vals)
    stackbounds = list(chain(*bounds))

    return stackvalue, stackbounds


def shape(parameter):
    """
    Get the shape of a ``Parameter``.

    Parameters
    ----------
    parameter: Parameter
        ``Parameter`` object to get the shape of

    Returns
    -------
    tuple:
        shape of the ``Parameter`` object
    """
    return parameter.shape
